Require Import Coq.Bool.Bool.
Require Import Coq.Arith.Arith.
Require Import Coq.Lists.List.
Require Import Coq.Program.Equality.
Require Import Coq.omega.Omega.
Require Export lattice.

Module LabelMap.


Import ListNotations.
Import SecurityLattice.

(* App id *)
Definition aid : Type := nat.

Definition beq_aid := beq_nat.

(* Instance id *)
Inductive iid : Type :=  
  | Iid : nat -> iid.

Definition beq_iid ii1 ii2 :=
  match ii1,ii2 with
    | Iid n1, Iid n2 => beq_nat n1 n2
  end.

Inductive app_label_map : Type :=
 | Single_Instance_App_Unlaunched (sl :label) (prev_ii : iid)
 | Single_Instance_App_Launched (sl : label) (dl : label) (ii: iid)
 | Multi_Instance_App (sl : label) (dl_list : list (iid * label))
 | NA.

Definition default_alm := Single_Instance_App_Unlaunched TOP (Iid 0).

(* The Label Map maps apps and instances into labels *)
Definition label_map : Type := list app_label_map.


(** Utility Functions *)
Fixpoint update_label_map_app (lm : label_map) (ai:aid) (rep_alm : app_label_map): label_map :=
  match lm with
  | nil => nil
  | alm :: tl => match ai with
                     | O => rep_alm :: tl
                     | S n => alm :: (update_label_map_app tl n rep_alm)
                     end
  end.


Fixpoint next_iid_nat (dll : list (iid * label)) : nat := 
match dll with
  | nil => O
  | (Iid n,_) :: tl => S (max n (next_iid_nat tl))
end.
Definition next_iid (dll : list (iid * label)) : iid := Iid (next_iid_nat dll).
Definition increment_iid (ii : iid) : iid := match ii with | Iid n => Iid (n+1) end.

Definition get_app_label_map (lm: label_map) (ai: aid) : option app_label_map :=
  nth_error lm ai.

Definition get_inst_label (dll : list (iid * label)) (ii:iid) : option label :=
option_map snd (find (fun a => beq_iid (fst a) ii) dll).

Definition get_label (alm: app_label_map) (ii: iid) : option label :=
  match alm with
  | Single_Instance_App_Launched sl dl ii2 => if beq_iid ii ii2 then Some dl else None
  | Multi_Instance_App sl dll => get_inst_label  dll ii
  | _ => None
  end.

(** Label Map Operations**)

Inductive has_app_label_map :
  label_map ->
  aid ->
  app_label_map -> Prop :=
  | halm_next lm ai alm alm2:
     has_app_label_map lm ai alm ->
     has_app_label_map ( alm2 :: lm ) (S ai) alm
  | halm_get lm alm:
     has_app_label_map ( alm :: lm ) O alm.

Inductive has_inst_label:
  list (iid * label) ->
  iid -> option label -> Prop :=
  | hil_next dll ii l ii2 l2:
    beq_iid ii ii2 = false ->
    has_inst_label dll ii l ->
    has_inst_label  ((ii2,l2) :: dll) ii l
  | hil_nil ii: has_inst_label nil ii None
  | hil_get dll ii l:
      has_inst_label ( (ii,l) :: dll) ii (Some l).

Inductive has_label:
  label_map ->
  aid -> iid ->
  option (label * label) -> Prop :=
  | hl_sl lm ai ii sl dl:
    has_app_label_map lm ai (Single_Instance_App_Launched sl dl ii) ->
    has_label lm ai ii (Some (sl,dl))
  | hl_sl_none lm ai ii ii' sl dl:
    has_app_label_map lm ai (Single_Instance_App_Launched sl dl ii') ->
    beq_iid ii ii' = false ->
    has_label lm ai ii None
  | hl_m lm ai ii sl dll dl:
    has_app_label_map lm ai (Multi_Instance_App sl dll) ->
    has_inst_label dll ii (Some dl) ->
    has_label lm ai ii (Some (sl,dl))
  | hl_m_none lm ai ii sl dll:
    has_app_label_map lm ai (Multi_Instance_App sl dll) ->
    has_inst_label dll ii None ->
    has_label lm ai ii None.


Inductive update_inst_label:
  list (iid * label) ->
  iid -> label -> 
  list (iid * label) -> Prop :=
  | uil_next dll ii l ii2 l2 dll2:
    beq_iid ii ii2 = false ->
    update_inst_label dll ii l dll2 ->
    update_inst_label  ((ii2,l2) :: dll) ii l ((ii2,l2) :: dll2)
  | uil_update dll ii l l_old:
      update_inst_label ( (ii,l_old) :: dll) ii l ( (ii,l) :: dll).

Inductive update_app_label_map :
  label_map ->
  aid ->
  app_label_map ->
  label_map -> Prop :=
  | ualm_next lm lm2 ai alm alm2:
     update_app_label_map lm ai alm lm2 ->
     update_app_label_map ( alm2 :: lm ) (S ai) alm ( alm2 :: lm2)
  | ualm_get lm alm alm_old:
     update_app_label_map ( alm_old :: lm ) O alm (alm :: lm).


End LabelMap.
